@@ -4,7 +4,7 @@
 
 import copy
 import typing
-from typing import Any, Mapping
+from typing import Any, Dict, Mapping, Optional
 
 PARAMETERS_STR = "$parameters"
 
@@ -94,7 +94,8 @@ class ManifestComponentTransformer:
         parent_field_identifier: str,
         declarative_component: Mapping[str, Any],
         parent_parameters: Mapping[str, Any],
-    ) -> Mapping[str, Any]:
+        use_parent_parameters: Optional[bool] = None,
+    ) -> Dict[str, Any]:
         """
         Recursively transforms the specified declarative component and subcomponents to propagate parameters and insert the
         default component type if it was not already present. The resulting transformed components are a deep copy of the input
@@ -103,6 +104,7 @@ class ManifestComponentTransformer:
         :param declarative_component: The current component that is having type and parameters added
         :param parent_field_identifier: The name of the field of the current component coming from the parent component
         :param parent_parameters: The parameters set on parent components defined before the current component
+        :param use_parent_parameters: If set, parent parameters will be used as the source of truth when key names are the same
         :return: A deep copy of the transformed component with types and parameters persisted to it
         """
         propagated_component = dict(copy.deepcopy(declarative_component))
@@ -118,19 +120,36 @@ class ManifestComponentTransformer:
             if found_type:
                 propagated_component["type"] = found_type
 
+        # Combines parameters defined at the current level with parameters from parent components. Parameters at the current
+        # level take precedence
+        current_parameters = dict(copy.deepcopy(parent_parameters))
+        component_parameters = propagated_component.pop(PARAMETERS_STR, {})
+        current_parameters = (
+            {**component_parameters, **current_parameters}
+            if use_parent_parameters
+            else {**current_parameters, **component_parameters}
+        )
+
         # When there is no resolved type, we're not processing a component (likely a regular object) and don't need to propagate parameters
         # When the type refers to a json schema, we're not processing a component as well. This check is currently imperfect as there could
         # be json_schema are not objects but we believe this is not likely in our case because:
         # * records are Mapping so objects hence SchemaLoader root should be an object
         # * connection_specification is a Mapping
-        if "type" not in propagated_component or self._is_json_schema_object(propagated_component):
+        if self._is_json_schema_object(propagated_component):
             return propagated_component
 
-        # Combines parameters defined at the current level with parameters from parent components. Parameters at the current
-        # level take precedence
-        current_parameters = dict(copy.deepcopy(parent_parameters))
-        component_parameters = propagated_component.pop(PARAMETERS_STR, {})
-        current_parameters = {**current_parameters, **component_parameters}
+        # For objects that don't have type check if their object fields have nested components which should have `$parameters` in it.
+        # For example, QueryProperties in requester.request_parameters, etc.
+        # Update propagated_component value with nested components with parent `$parameters` if needed and return propagated_component.
+        if "type" not in propagated_component:
+            if self._has_nested_components(propagated_component):
+                propagated_component = self._process_nested_components(
+                    propagated_component,
+                    parent_field_identifier,
+                    current_parameters,
+                    use_parent_parameters,
+                )
+            return propagated_component
 
         # Parameters should be applied to the current component fields with the existing field taking precedence over parameters if
         # both exist
@@ -145,7 +164,10 @@ class ManifestComponentTransformer:
                 excluded_parameter = current_parameters.pop(field_name, None)
                 parent_type_field_identifier = f"{propagated_component.get('type')}.{field_name}"
                 propagated_component[field_name] = self.propagate_types_and_parameters(
-                    parent_type_field_identifier, field_value, current_parameters
+                    parent_type_field_identifier,
+                    field_value,
+                    current_parameters,
+                    use_parent_parameters=use_parent_parameters,
                 )
                 if excluded_parameter:
                     current_parameters[field_name] = excluded_parameter
@@ -158,7 +180,10 @@ class ManifestComponentTransformer:
                             f"{propagated_component.get('type')}.{field_name}"
                         )
                         field_value[i] = self.propagate_types_and_parameters(
-                            parent_type_field_identifier, element, current_parameters
+                            parent_type_field_identifier,
+                            element,
+                            current_parameters,
+                            use_parent_parameters=use_parent_parameters,
                         )
                 if excluded_parameter:
                     current_parameters[field_name] = excluded_parameter
@@ -169,4 +194,33 @@ class ManifestComponentTransformer:
 
     @staticmethod
     def _is_json_schema_object(propagated_component: Mapping[str, Any]) -> bool:
-        return propagated_component.get("type") == "object"
+        return propagated_component.get("type") == "object" or propagated_component.get("type") == [
+            "null",
+            "object",
+        ]
+
+    @staticmethod
+    def _has_nested_components(propagated_component: Dict[str, Any]) -> bool:
+        for k, v in propagated_component.items():
+            if isinstance(v, dict) and v.get("type"):
+                return True
+        return False
+
+    def _process_nested_components(
+        self,
+        propagated_component: Dict[str, Any],
+        parent_field_identifier: str,
+        current_parameters: Mapping[str, Any],
+        use_parent_parameters: Optional[bool] = None,
+    ) -> Dict[str, Any]:
+        for field_name, field_value in propagated_component.items():
+            if isinstance(field_value, dict) and field_value.get("type"):
+                nested_component_with_parameters = self.propagate_types_and_parameters(
+                    parent_field_identifier,
+                    field_value,
+                    current_parameters,
+                    use_parent_parameters=use_parent_parameters,
+                )
+                propagated_component[field_name] = nested_component_with_parameters
+
+        return propagated_component
