@@ -3,8 +3,23 @@
 #
 
 from airbyte_cdk.sources.declarative.schema.default_schema_loader import DefaultSchemaLoader
+from airbyte_cdk.sources.declarative.schema.dynamic_schema_loader import (
+    ComplexFieldType,
+    DynamicSchemaLoader,
+    SchemaTypeIdentifier,
+    TypesMap,
+)
 from airbyte_cdk.sources.declarative.schema.inline_schema_loader import InlineSchemaLoader
 from airbyte_cdk.sources.declarative.schema.json_file_schema_loader import JsonFileSchemaLoader
 from airbyte_cdk.sources.declarative.schema.schema_loader import SchemaLoader
 
-__all__ = ["JsonFileSchemaLoader", "DefaultSchemaLoader", "SchemaLoader", "InlineSchemaLoader"]
+__all__ = [
+    "JsonFileSchemaLoader",
+    "DefaultSchemaLoader",
+    "SchemaLoader",
+    "InlineSchemaLoader",
+    "DynamicSchemaLoader",
+    "ComplexFieldType",
+    "TypesMap",
+    "SchemaTypeIdentifier",
+]
